@@ -3,12 +3,14 @@ import sys
 import time
 import queue
 import socket
+import unittest
 import binascii
 import multiprocessing
 
 from .const import *
 from .misc import example_bytes,_x,chunk,dattr
 from .blocks import *
+import m17.network as network
 
 def default_config(c2_mode):
     c2,conrate,bitframe = codec2setup(c2_mode)
@@ -30,6 +32,85 @@ def default_config(c2_mode):
         })
     return config
 
+
+def m17_parrot(refcallsign, port=default_port):
+    #A parrot service for M17 - it's a full client that records and plays back after incoming stream is over PTT is released
+    port=int(port)
+    # udp_recv and udp_send are needed here, too.
+    ...
+
+def m17_mirror(refcallsign, port=default_port):
+    #reflects your M17 stream back to you after decoding and encoding
+    #can be useful for later transformations, like testing voice stream compatibilites
+    port=int(port)
+    # udp_recv and udp_send are needed here, too.
+    ...
+
+def udp_mirror(refcallsign, port=default_port):
+    # reflects your own UDP packets back to you after a delay
+    port=int(port)
+
+    pkts = {}
+    def packet_handler(sock, active_connections, bs, conn):
+        if conn not in pkts:
+            pkts[ conn ] = dattr({ "packets":[], "lastseen":time.time()})
+        else:
+            this = pkts[conn]
+            pkts[conn].packets.append( (time.time()-this.lastseen, bs ) )
+            pkts[conn].lastseen = time.time()
+    def timer(sock):
+        def replay(conn,packets):
+            for reltime,bs in packets:
+                time.sleep(reltime)
+                sock.sendto( bs, conn) 
+        delthese = []
+        for conn in pkts:
+            if pkts[conn].lastseen + 10 < time.time():
+                #as udp_server is written, this will stop us from recvfrom - and that's okay for now
+                #if we have multiple users, we may well timeout on several in a row because of the delays we're seeing here
+                #what i wish i had was a setTimeout like in JS, but I'm sure I can do something with asyncio later to get what I want (and actually support multiple udp_mirror users)
+                replay(conn, pkts[conn].packets) 
+                delthese.append(conn) 
+        for conn in delthese:
+            del pkts[conn]
+    srv = udp_server(port, packet_handler, timer)
+    srv()
+
+def udp_reflector(refcallsign, port=default_port):
+    # "Reflects" an incoming stream to all connected users.
+    # ✔ So first, we need a way to receive connections and keep track of them, right?
+    # We also have our own callsign, but we'll deal with that later.
+
+    port=int(port)
+    def packet_handler(sock, active_connections, bs, conn):
+        others = [c for c in active_connections.keys() if c != conn]
+        for c in others:
+            sock.sendto(bs, c)
+    srv = udp_server(port, packet_handler)
+    srv()
+
+
+def m17ref_client(mycall,mymodule,refname,module,port=default_port,mode=3200):
+    mode=int(mode) #so we can call modular_client straight from command line
+    port=int(port)
+    if( refname.startswith("M17-") and len(refname) <= 7 ):
+        #should also be able to look up registered port in dns
+        host = network.m17ref_name2host(refname)
+        print(host)
+        #fallback to fetching json if its not in dns already
+    else:
+        raise(NotImplementedError)
+    myrefmod = "%s %s"%(mycall,mymodule)
+    c = m17ref_client_blocks(myrefmod,module,host,port)
+    tx_chain = [mic_audio, codec2enc, vox, m17frame, tobytes, c.sender()]
+    rx_chain = [c.receiver(), m17parse, payload2codec2, codec2dec, spkr_audio]
+    config = default_config(mode)
+    config.m17.dst = "%s %s"%(refname,module)
+    config.m17.src = mycall
+    print(config)
+    c.start()
+    modular(config, [tx_chain, rx_chain])
+
 def voipsim(host="localhost",src="W2FBI",dst="SP5WWP",mode=3200,port=default_port):
     mode=int(mode) #so we can call modular_client straight from command line
     port=int(port)
@@ -64,7 +145,7 @@ def recv_dump(mode=3200,port=default_port):
     config = default_config(mode)
     modular(config, [rx_chain])
 
-def voip(host="localhost",src="W2FBI",dst="SP5WWP",mode=3200,port=default_port, voipmode="full"):
+def voip(host="localhost",port=default_port,voipmode="full",mode=3200,src="W2FBI",dst="SP5WWP"):
     mode=int(mode) #so we can call modular_client straight from command line
     port=int(port)
 
@@ -92,39 +173,31 @@ def voip(host="localhost",src="W2FBI",dst="SP5WWP",mode=3200,port=default_port,
 
     modular(config, [tx_chain, rx_chain])
 
-def _echolink_bridge():
-    #this is incomplete, and must remain so until a proper reflector enters the picture
-    #it's here mostly to save it for later
-    echolink_bridge_in = [udp_recv(55501), chunker_b(640), np_convert("<h"), integer_decimate(2), codec2enc, m17frame, tobytes, udp_send((host,port)) ]
-    m17_to_echolink = [
-            udp_recv(17000), tee("rx"), 
-            m17parse, payload2codec2, codec2dec, 
-            integer_interpolate(2),
-            udp_send(("pidp8i",55500)) 
-            ]
-    echolink_bridge_monitor = [
-                udp_recv(55501), 
-                chunker_b(640), 
-                np_convert("<h"), 
-                integer_decimate(2), 
-                codec2enc, 
-                tee("RX"),
-                m17frame, tobytes,
-                m17parse,payload2codec2, 
-                tee("RX"),
-                codec2dec, 
-                spkr_audio
-                ]
-    echolink_bridge_out = [
-            mic_audio, codec2enc, vox, m17frame, tobytes, 
-            m17parse, payload2codec2, codec2dec, 
-            integer_interpolate(2),
-            udp_send(("pidp8i",55500)) 
-            ]
+def echolink_bridge(mycall,mymodule,refname,refmodule,refport=default_port,mode=3200):
+    mode=int(mode) #so we can call modular_client straight from command line
+    refport=int(port)
+    if( refname.startswith("M17-") and len(refname) <= 7 ):
+        #should also be able to look up registered port in dns
+        host = network.m17ref_name2host(refname)
+        print(host)
+        #fallback to fetching json if its not in dns already
+    else:
+        raise(NotImplementedError)
+    myrefmod = "%s %s"%(mycall,mymodule)
+    c = m17ref_client_blocks(myrefmod,refmodule,host,refport)
+    echolink_to_m17ref = [udp_recv(55501), chunker_b(640), np_convert("<h"), integer_decimate(2), codec2enc, m17frame, tobytes, c.sender()]
+    m17ref_to_echolink = [ c.receiver(), m17parse, payload2codec2, codec2dec, integer_interpolate(2), udp_send(("127.0.0.1",55500)) ]
     config = default_config(mode)
+    config.m17.dst = "%s %s"%(refname,refmodule)
+    config.m17.src = mycall
+    print(config)
+    c.start()
     modular(config, [tx_chain, rx_chain])
 
 def m17_to_echolink(port=default_port, echolink_host="localhost",mode=3200, echolink_audio_in_port=55500):
+    port=int(port)
+    mode=int(mode)
+    echolink_audio_in_port=int(echolink_audio_in_port)
     """
     decode and bridge m17 packets to echolink
     (useful for interopability testing)
